// richedView.cpp : implementation of the CRichSyntaxView class
//

#include "stdafx.h"
//#include "richedView.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

// File interaction, used in OnSave and OnOpen respectivly
DWORD CALLBACK WriteFileCallBack(DWORD dwCookie, LPBYTE pbBuff, LONG cb, LONG FAR* pcb);
DWORD CALLBACK ReadFileCallBack(DWORD dwCookie, LPBYTE pbBuff, LONG cb, LONG FAR* pcb);
/////////////////////////////////////////////////////////////////////////////
// CRichSyntaxView

IMPLEMENT_DYNCREATE(CRichSyntaxView, CRichEditView)

BEGIN_MESSAGE_MAP(CRichSyntaxView, CRichEditView)
	//{{AFX_MSG_MAP(CRichSyntaxView)
	ON_WM_DESTROY()
	ON_WM_CHAR()
	ON_WM_KEYUP()
	ON_WM_LBUTTONDOWN()
	//}}AFX_MSG_MAP
	// Standard printing commands
	ON_COMMAND(ID_FILE_PRINT, CRichEditView::OnFilePrint)
	ON_COMMAND(ID_EDIT_PASTE, OnPaste)
	ON_COMMAND(ID_FILE_PRINT_DIRECT, CRichEditView::OnFilePrint)
	ON_COMMAND(ID_FILE_PRINT_PREVIEW, CRichEditView::OnFilePrintPreview)
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CRichSyntaxView construction/destruction

CRichSyntaxView::CRichSyntaxView()
{
	USER_DEFINED = RGB(100, 100, 0);  //Color for AddUserDefined words
	KEYWORD = RGB(0, 0, 200); //Color for AddKeyWord words
	COMMENT = RGB(0, 200, 0); //Color for AddComment words/characters
	bCaseSensitive = true; // if this is case sensitive(Compare or CompareNoCase is used)
	iLastWord[0] = 0; // setup int defaults
	iLastWord[1] = 0; // see OnKeyUp and OnChar for why these are here
	iLastComment[0] = 0;
	iLastComment[1] = 0;
}

CRichSyntaxView::CRichSyntaxView(bool caseSensitive, COLORREF keyword, COLORREF userword, COLORREF comment)
{
	// See comments in default constructor for explenation
	USER_DEFINED = userword;
	KEYWORD = keyword;
	COMMENT = comment;
	bCaseSensitive = caseSensitive;
	iLastWord[0] = 0;
	iLastWord[1] = 0;
	iLastComment[0] = 0;
	iLastComment[1] = 0;
}

CRichSyntaxView::~CRichSyntaxView()
{
}

void CRichSyntaxView::OnInitialUpdate()
{
	CRichEditView::OnInitialUpdate();
	CHARFORMAT cf;
	cf.dwMask = CFM_FACE | CFM_BOLD;
	cf.dwEffects = 0;
	strcpy(cf.szFaceName, "Arial");
	GetRichEditCtrl().SetDefaultCharFormat(cf);

	// Set the printing margins (720 twips = 1/2 inch).
	SetMargins(CRect(720, 720, 720, 720));
}

/////////////////////////////////////////////////////////////////////////////
// CRichSyntaxView printing

BOOL CRichSyntaxView::OnPreparePrinting(CPrintInfo* pInfo)
{
	// default preparation
	return DoPreparePrinting(pInfo);
}


void CRichSyntaxView::OnDestroy()
{
	// Deactivate the item on destruction; this is important
	// when a splitter view is being used.
   CRichEditView::OnDestroy();
   COleClientItem* pActiveItem = GetDocument()->GetInPlaceActiveItem(this);
   if (pActiveItem != NULL && pActiveItem->GetActiveView() == this)
   {
      pActiveItem->Deactivate();
      ASSERT(GetDocument()->GetInPlaceActiveItem(this) == NULL);
   }
}


/////////////////////////////////////////////////////////////////////////////
// CRichSyntaxView diagnostics

#ifdef _DEBUG
void CRichSyntaxView::AssertValid() const
{
	CRichEditView::AssertValid();
}

void CRichSyntaxView::Dump(CDumpContext& dc) const
{
	CRichEditView::Dump(dc);
}

#endif //_DEBUG

/////////////////////////////////////////////////////////////////////////////
// CRichSyntaxView message handlers

void CRichSyntaxView::OnPaste()
{
	GetRichEditCtrl().Paste();
	Parse();
}

void CRichSyntaxView::OnLButtonDown(UINT nFlags, CPoint point) 
{
	CRichEditView::OnLButtonDown(nFlags, point);
	FindComment();	// since we land someplace fairly random figure out where the nearist comment
	FindWord();		// and full word are
}

void CRichSyntaxView::OnKeyUp(UINT nChar, UINT nRepCnt, UINT nFlags) 
{
	CRichEditView::OnKeyDown(nChar, nRepCnt, nFlags);
	switch(nChar) {	
	case VK_UP:
	case VK_DOWN:
	case VK_RIGHT:
	case VK_LEFT:  // see OnLButtonDown
		FindComment();
		FindWord();
		break;
	case VK_DELETE: // delete logic similar to backspace logic except in the other direction
		CHARRANGE cr;

		FindComment();
		FindWord();
		GetRichEditCtrl().GetSel(cr);
		
		if(iLastComment[0] == cr.cpMin || iLastComment[0] == (cr.cpMin - 1))
			ParseRange(iLastComment[0], -1);
		else if(iLastComment[1] == cr.cpMin || iLastComment[1] == (cr.cpMin + 1))
			ParseRange(iLastComment[1], -1);
		else if((iLastWord[0] >= (cr.cpMin) && iLastWord[1] <= cr.cpMin) 
				|| (iLastWord[1] <= cr.cpMin - 2))
			ParseRange(iLastWord[0], iLastWord[1]);
		else if(iLastComment[0] == cr.cpMax || iLastComment[0] == (cr.cpMax - 1))
			ParseRange(iLastComment[0], -1);
		else if(iLastComment[1] == cr.cpMax || iLastComment[1] == (cr.cpMax + 1))
			ParseRange(iLastComment[1], -1);
		else if((iLastWord[0] >= (cr.cpMax) && iLastWord[1] <= cr.cpMax) 
				|| (iLastWord[1] <= cr.cpMax - 2))
			ParseRange(iLastWord[0], iLastWord[1]);
		break;
	}
}

void CRichSyntaxView::OnChar(UINT nChar, UINT nRepCnt, UINT nFlags) 
{
	CRichEditView::OnChar(nChar, nRepCnt, nFlags);
	CHARRANGE cr;
	int iEnd;
	CRichEditCtrl& r = GetRichEditCtrl();

	r.GetSel(cr);
	iEnd = cr.cpMin;

	switch(nChar) {
	case VK_RETURN: 
		FindWord();
		if(iLastComment[0] <= iEnd && (iLastComment[1] >= (iEnd - 1) || iLastComment[1] == -1))
			ParseRange(iLastComment[0], -1);
		else if(iLastWord[0] < iEnd && iLastWord[1] > iEnd)
			ParseRange(iLastWord[0], -1); // if breaks up last word
		break;
	case VK_BACK:
		FindComment();
		FindWord();
 		if(iLastComment[0] == iEnd || iLastComment[0] == (iEnd - 1))
			ParseRange(iLastComment[0], -1); // if destroys the beginning of a MLine comment
		else if(iLastComment[1] == (iEnd + 1) || iLastComment[1] == iEnd) {
			ResetRange(iLastComment[1], -1);
			ParseRange(iLastComment[0], -1); // if destroys the end of a MLine comment
		}
		else if(iLastWord[0] <= (iEnd + 1) && iLastWord[1] >= (iEnd + 1))
			ParseRange(iLastWord[0], iLastWord[1]); // if destroys last word
		break;
	case int(' '):
		if(iLastComment[0] == (iEnd + 1))
			ParseRange(iLastComment[0], -1); // if destroys the beginning of a MLine comment
		else if(iLastComment[1] == (iEnd))
			ParseRange(iLastComment[0], -1); // if destroys the end of a MLine comment
		else if(!CompareFormat(iLastWord[0], iLastWord[1]+1))
			ParseRange(iLastWord[0], iLastWord[1]+1); // if adding a space changes the layout
		if(iLastWord[1] < iEnd) {
			iLastWord[0] = iEnd; // set up last word to quoridates of space
			iLastWord[1]++;
		}
		break;
	default:
		FindWord();
		if(iLastComment[0] < (iEnd - 1) && (iLastComment[1] > iEnd || iLastComment[1] == -1)) {
			if(iLastComment[1] != -1)
				 iLastComment[1] += 1;
			ParseRange(iLastComment[0], iLastComment[1]); // if within a multi line comment
		}
		else if(iLastWord[0] <= (iEnd - 1) && iLastWord[1] >= iEnd)
			ParseRange(iLastWord[0], iLastWord[1]); // if within the last word
		else if((iLastWord[1] - iLastWord[0]) > 0) 
			ParseRange(iLastWord[0], iLastWord[1]); // if last word wasn't a space
		else if(!CompareFormat(iLastWord[0], iEnd)) 
			ParseRange(iLastWord[0] - 1, iLastWord[1]); // Compare format decides is it would look the same w/ out formating
		else
			ParseRange(iLastWord[0], iEnd); // if all else fails
	}
}

//////////////////////////////////////////////////////
// Syntax logic

void CRichSyntaxView::Parse()
{
	CRichEditCtrl& r = GetRichEditCtrl();
	int iCount, iCCount, iSize;
	int iWordStart = 0;
	int iWordEnd = 0;
	CString word, cLine;
	CHARFORMAT cf;
	CHARRANGE ccr;

	cf.dwMask = CFM_BOLD;
	cf.dwEffects = 0;
	iSize = r.GetTextLength();
	r.GetWindowText(cLine);

	r.GetSel(ccr);
	iCount = 0;

SetRedraw(FALSE); //	LockWindowUpdate();
//	r.HideSelection(TRUE, FALSE);
	while(iCount < iSize) {
		word.Empty();
		while(iCount < iSize && IsWhiteSpace(cLine[iCount]))
			iCount++;
		iWordStart = iCount;
	
		while(iCount < iSize && !IsWhiteSpace(cLine[iCount])) {
			word += cLine[iCount];
			iCount++;
		}
		iWordEnd = iCount;

		// Select correct formating for word */
		// Comments(ie // int C or '' in VB run untill end of line
		if(IsComment(word.Left(2))) {
			while(iCount < iSize) {
				if(cLine[iCount] == '\n')
					break;
				iCount++;
			}
			iLastComment[0] = iWordStart;
			iLastComment[1] = -1;

			r.SetSel(iWordStart, iCount);
			cf.dwMask |= CFM_COLOR;
			cf.crTextColor = COMMENT;
			r.SetSelectionCharFormat(cf);
		}
		// this is a cheap hack for right now, hardcoding is bad
		else if(word.Left(2).CompareNoCase(_T("/*")) == 0) {
			cf.dwMask |= CFM_COLOR;
			cf.crTextColor = COMMENT;
			iLastComment[0] = iWordStart;
			iLastComment[1] = -1;
			iCCount = iWordStart + 1;
			while(iCCount < iSize) {
				if(cLine[iCCount - 1] == '*' && cLine[iCCount] == '/') {
					iLastComment[1] = iCCount;
					break;
				}
				iCCount++;
			}
			iCount = ++iCCount;
			r.SetSel(iWordStart, iCount);
			r.SetSelectionCharFormat(cf);
		}
		else if(IsKeyWord(word)) {
			r.SetSel(iWordStart, iWordEnd);
			cf.dwMask |= CFM_COLOR;
			cf.crTextColor = KEYWORD;
			r.SetSelectionCharFormat(cf);
		}
		else if(IsUserDefined(word)) {
			r.SetSel(iWordStart, iWordEnd);
			cf.dwMask |= CFM_COLOR;
			cf.crTextColor = USER_DEFINED;
			r.SetSelectionCharFormat(cf);
		}
	}

	r.SetSel(ccr);
	cf.dwMask = CFM_COLOR;
	cf.crTextColor = GetSysColor(COLOR_WINDOWTEXT);
	r.SetSelectionCharFormat(cf);

	iLastWord[0] = iWordStart;
	iLastWord[1] = iWordEnd;
//	r.HideSelection(FALSE, FALSE);
SetRedraw(TRUE); //	UnlockWindowUpdate();
Invalidate(FALSE);
}

void CRichSyntaxView::ParseRange(int iMin, int iMax)
{
	CRichEditCtrl& r = GetRichEditCtrl();
	int iCount;
	int iWordStart = 0;
	int iWordEnd = 0;
	if(iMax == -1)
		iMax = r.GetTextLength();
	int iSize;
	CString word, cLine;
	CHARFORMAT cf;
	CHARRANGE ccr;

	if(iMax < iMin) {
		{
			int iTmp = iMax;
			iMax = iMin;
			iMin = iTmp;
		}
	}
	cf.dwMask = CFM_BOLD;
	cf.dwEffects = 0;
	iCount = 0;

SetRedraw(FALSE); //	LockWindowUpdate();
	r.HideSelection(TRUE, FALSE);
	r.GetSel(ccr);
	r.SetSel(iMin, iMax);
	// normilize the selection
	cf.dwMask |= CFM_COLOR;
	cf.crTextColor = GetSysColor(COLOR_WINDOWTEXT);
//	r.SetSelectionCharFormat(cf);
	r.SetSelectionCharFormat(cf);

	cLine = r.GetSelText();
	iSize = cLine.GetLength();
	while(iCount < iSize) {
		word.Empty();
		while(iCount < iSize && IsWhiteSpace(cLine[iCount]))
			iCount++;
		iWordStart = iCount;

		while(iCount < iSize && !IsWhiteSpace(cLine[iCount])) {
			word += cLine[iCount];
			iCount++;
		}
		iWordEnd = iCount;

		// Select correct formating for word
		if(IsComment(word.Left(2))) {
			iLastComment[0] = iWordStart + iMin;
			iLastComment[1] = -1;

			while(iCount < (iSize)) {
				if(cLine[iCount] == 13)
					iLastComment[1] = iCount + iMin;
				iCount++;
			}

			r.SetSel(iWordStart + iMin, (iCount + 1) + iMin);
			cf.dwMask |= CFM_COLOR;
			cf.crTextColor = COMMENT;
			r.SetSelectionCharFormat(cf);
			break;
		}
		
		else if(word.Left(2).CompareNoCase(_T("/*")) == 0) {
			cf.dwMask |= CFM_COLOR;
			cf.crTextColor = COMMENT;
			iCount++;
			iLastComment[0] = iWordStart + iMin;
			iLastComment[1] = -1;
			while(iCount < iSize) {
				if(cLine[iCount - 1] == '*' && cLine[iCount] == '/') {
					iCount++;
					iLastComment[1] = iCount + iMin;
					break;
				}
				iCount++;
			}
			r.SetSel(iWordStart + iMin, iCount + iMin);
			r.SetSelectionCharFormat(cf);
		}
		else if(IsKeyWord(word)) {
			r.SetSel(iWordStart + iMin, iWordEnd + iMin);
			cf.dwMask |= CFM_COLOR;
			cf.crTextColor = KEYWORD;
			r.SetSelectionCharFormat(cf);
		}
		else if(IsUserDefined(word)) {
			r.SetSel(iWordStart + iMin, iWordEnd + iMin);
			cf.dwMask |= CFM_COLOR;
			cf.crTextColor = USER_DEFINED;
			r.SetSelectionCharFormat(cf);
		}
	}

	r.SetSel(ccr);
	cf.dwMask = CFM_COLOR;
	cf.crTextColor = GetSysColor(COLOR_WINDOWTEXT);
	r.SetSelectionCharFormat(cf);

	iLastWord[0] = iWordStart + iMin;
	iLastWord[1] = iWordEnd + iMin;
	r.HideSelection(FALSE, FALSE);
SetRedraw(TRUE); //	UnlockWindowUpdate();
Invalidate(FALSE);
}

/*	this is a test to help reduce flickering in MLine comments 
	Not currently implemented
*/
void CRichSyntaxView::ParseComment(int iMin, int iMax)
{
	CRichEditCtrl& r = GetRichEditCtrl();
	CHARFORMAT cf;
	CHARRANGE ccr;

	cf.dwMask = CFM_BOLD;
	cf.dwEffects = 0;

	LockWindowUpdate();
	r.HideSelection(TRUE, FALSE);
	r.GetSel(ccr);
	r.SetSel(iMin, iMax);

	cf.dwMask |= CFM_COLOR;
	cf.crTextColor = COMMENT;
	r.SetSel(iMin, iMax);
	r.SetSelectionCharFormat(cf);

	r.SetSel(ccr);
	cf.dwMask = CFM_COLOR;
	cf.crTextColor = GetSysColor(COLOR_WINDOWTEXT);
	r.SetSelectionCharFormat(cf);
	r.HideSelection(FALSE, FALSE);
	UnlockWindowUpdate();
}

bool CRichSyntaxView::IsWhiteSpace(char c)
{
	if(c == ' ' || c == 13 || c == 10 || c == '\t')
		return true;
	return false;
}

bool CRichSyntaxView::IsUserDefined(CString word)
{
	POSITION p;
	p = userDefined.GetHeadPosition();
	while(p) {
		if(bCaseSensitive)  {
			if(word.Compare(userDefined.GetNext(p)) == 0)
				return true;
		}
		else {
			if(word.CompareNoCase(userDefined.GetNext(p)) == 0)
				return true;
		}
	}
	return false;
}

bool CRichSyntaxView::IsKeyWord(CString word)
{
	POSITION p;
	p = keyWords.GetHeadPosition();
	while(p) {
		if(bCaseSensitive) {
			if(word.Compare(keyWords.GetNext(p)) == 0)
				return true;
		}
		else {
			if(word.CompareNoCase(keyWords.GetNext(p)) == 0)
				return true;
		}
	}
	return false;
}

bool CRichSyntaxView::IsComment(CString word)
{
	POSITION p;
	p = comment.GetHeadPosition();
	while(p) {
		if(bCaseSensitive) {
			if(word.Compare(comment.GetNext(p)) == 0)
				return true;
		}
		else {
			if(word.CompareNoCase(comment.GetNext(p)) == 0)
				return true;
		}
	}
	return false;
}

bool CRichSyntaxView::LoadKeyWordFile(CString word)
{
	CStdioFile file;
	file.Open(word, CFile::modeRead | CFile::shareDenyWrite);
	LoadKeyWordFile(file);
	return true;
}

bool CRichSyntaxView::LoadKeyWordFile(CStdioFile& file)
{
	CString sLine;

	while(file.ReadString(sLine)) {
		if(sLine[0] == '&') {
			userDefined.AddTail(sLine.Right(sLine.GetLength() - 1));
		}
		else if(sLine[0] == '!') {
			comment.AddTail(sLine.Right(sLine.GetLength() - 1));
		}
		else {
			keyWords.AddTail(sLine.Right(sLine.GetLength() - 1));
		}
	}
	return true;
}

void CRichSyntaxView::AddKeyWord(LPCTSTR key)
{
	keyWords.AddTail(key);
}

void CRichSyntaxView::AddUserDefined(LPCTSTR key)
{
	userDefined.AddTail(key);
}

void CRichSyntaxView::AddComment(LPCTSTR key)
{
	comment.AddTail(key);
}

bool CRichSyntaxView::CompareFormat(int iMin, int iMax)
{
	int iCount, iWordStart, iWordEnd;
	int iSize = iMax - iMin;
	CRichEditCtrl& r = GetRichEditCtrl();
	CString word, cLine;
	CHARRANGE ccr;
	CHARFORMAT cf, ccf;

	cf.dwMask = CFM_BOLD;
	cf.dwEffects = 0;

	r.HideSelection(TRUE, FALSE);
	r.GetSel(ccr);
	r.SetSel(iMin, iMax);
	r.GetSelectionCharFormat(ccf);
		
	word.Empty();
	iCount = 0;
	cLine = r.GetSelText();
	while(iCount < iSize && IsWhiteSpace(cLine[iCount]))
		iCount++;
	iWordStart = iCount;
	while(iCount < iSize && !IsWhiteSpace(cLine[iCount])) {
		word += cLine[iCount];
		iCount++;
	}
	iWordEnd = iCount;


	if(IsKeyWord(word)) {
		cf.dwMask |= CFM_COLOR;
		cf.crTextColor = KEYWORD;
	}
	else if(IsComment(word)) {
		cf.dwMask |= CFM_COLOR;
		cf.crTextColor = COMMENT;
	}
	else if(IsUserDefined(word)) {
		cf.dwMask |= CFM_COLOR;
		cf.crTextColor = USER_DEFINED;
	}
	else {
		cf.dwMask = CFM_COLOR;
		cf.crTextColor = GetSysColor(COLOR_WINDOWTEXT);
	}
	
	if(cf.dwMask == ccf.dwMask && cf.crTextColor == ccf.crTextColor) {
		r.SetSel(ccr);
		r.HideSelection(FALSE, FALSE);
		return true;
	}

	r.SetSel(ccr);
	r.HideSelection(FALSE, FALSE);
	return true;
}

void CRichSyntaxView::ResetRange(int iMin, int iMax)
{
	CRichEditCtrl& r = GetRichEditCtrl();
	if(iMax == -1)
		iMax = r.GetTextLength();
	CHARFORMAT cf;
	CHARRANGE ccr;

	if(iMax < iMin) {
		{
			int iTmp = iMax;
			iMax = iMin;
			iMin = iTmp;
		}
	}
	cf.dwMask = CFM_BOLD;
	cf.dwEffects = 0;

	LockWindowUpdate();
	r.HideSelection(TRUE, FALSE);
	r.GetSel(ccr);
	r.SetSel(iMin, iMax);

	cf.dwMask = CFM_COLOR;
	cf.crTextColor = GetSysColor(COLOR_WINDOWTEXT);
	r.SetSel(iMin, iMax);
	r.SetSelectionCharFormat(cf);

	r.SetSel(ccr);
	r.HideSelection(FALSE, FALSE);
	UnlockWindowUpdate();
}


void CRichSyntaxView::FindComment()
{
	CRichEditCtrl& r = GetRichEditCtrl();
	int iCount, iSize, iWordStart, iWordEnd;
	CString word, cLine;

	r.GetWindowText(cLine);
	iSize = cLine.GetLength();
	iCount = 0;
	while(iCount < iSize) {
		word.Empty();
		while(iCount < iSize && IsWhiteSpace(cLine[iCount]))
			iCount++;
		iWordStart = iCount;

		while(iCount < iSize && !IsWhiteSpace(cLine[iCount])) {
			word += cLine[iCount];
			iCount++;
		}
		iWordEnd = iCount;
		if(word.Left(2).CompareNoCase(_T("/*")) == 0) {
			iLastComment[0] = iWordStart;
			iCount++;
			while(iCount < iSize) {
				if(cLine[iCount - 1] == _T('*') && cLine[iCount] == _T('/')) {
					iCount++;
					iLastComment[1] = iCount;// + 1;
					break;
				}
				iCount++;
			}
			break;
		}
		else if(IsComment(word.Left(2))) {
			iLastComment[0] = iWordStart;
			iLastComment[1] = -1;

			while(iCount < (iSize)) {
				if(cLine[iCount] == 13)
					iLastComment[1] = iCount;
				iCount++;
			}
		}
	}
}

void CRichSyntaxView::FindWord()
{
	CRichEditCtrl& r = GetRichEditCtrl();
	int iCount, iSize, iWordStart, iWordEnd;
	CString word, cLine;
	CHARRANGE ccr;

	r.GetSel(ccr);
	r.GetWindowText(cLine);
	iSize = cLine.GetLength();
	iCount = 0;
	while(iCount < iSize) {
		word.Empty();
		while(iCount < iSize && IsWhiteSpace(cLine[iCount]))
			iCount++;
		iWordStart = iCount;

		while(iCount < iSize && !IsWhiteSpace(cLine[iCount])) {
			word += cLine[iCount];
			iCount++;
		}
		iWordEnd = iCount;
		if(iWordStart <= ccr.cpMax && iWordEnd >= ccr.cpMax) {
			iLastWord[0] = iWordStart;
			iLastWord[1] = iWordEnd;
			break;
		}
	}
}

void CRichSyntaxView::SetFileName(CString word)
{
	sFileName = word;
}

void CRichSyntaxView::OnOpen()
{
	if(!sFileName.IsEmpty()) {
		CFile file(sFileName, CFile::modeRead | CFile::typeBinary);
		EDITSTREAM stream;
		stream.dwCookie = (DWORD)&file;
		stream.pfnCallback = ReadFileCallBack;
		GetRichEditCtrl().StreamIn(SF_RTF, stream);
	}
}

void CRichSyntaxView::OnSave()
{
	if(!sFileName.IsEmpty()) {
		CFile file(sFileName, CFile::modeCreate | CFile::modeWrite | CFile::typeBinary);
		EDITSTREAM stream;
		stream.dwCookie = (DWORD)&file;
		stream.pfnCallback = WriteFileCallBack;
		GetRichEditCtrl().StreamOut(SF_RTF, stream);
	}
}

DWORD CALLBACK WriteFileCallBack(DWORD dwCookie, LPBYTE pbBuff, LONG cb, LONG FAR* pcb)
{
	CFile *pFile = (CFile*)dwCookie;
	try
	{
		pFile->Write(pbBuff, cb);
	}
	catch(CFileException* pEX)
	{
		pEX;
		*pcb = 0;
		return 0;
	};

	*pcb = cb;
	return 1;
}

DWORD CALLBACK ReadFileCallBack(DWORD dwCookie, LPBYTE pbBuff, LONG cb, LONG FAR* pcb)
{
	CFile *pFile = (CFile*)dwCookie;
	try
	{
		pFile->Read(pbBuff, cb);
	}
	catch(CFileException* pEX)
	{
		pEX;
		*pcb = 0;
		return 0;
	};

	*pcb = cb;
	return 0;
}
